from typing import Tuple, Optional, Sequence

import numpy as np
import torch

from experiments.datasets.base import Array, DataProvider


class ConditionalBananaDataProvider(DataProvider):
    def __init__(
        self, k: int, d: int, seed: Optional[int] = 42, nonlinear: bool = False
    ):
        assert d in [2, 3, 4]
        assert k > 0
        super().__init__(seed=seed)
        self._d = d
        self._k = k
        self._beta = self._make_beta()
        self._nonlinear = nonlinear

    @property
    def k(self) -> int:
        return self._k

    @property
    def d(self) -> int:
        return self._d

    def sample_x(self, n: int) -> Array:
        X = self._rng.uniform(low=0.8, high=3.2, size=(n, self.k))
        return X

    def sample(self, n: int, x: Optional[Array] = None) -> Tuple[Array, Array]:
        if x is None:
            X = self.sample_x(n=n)
        else:
            assert len(x.shape) == 2
            assert x.shape[0] == 1
            X = np.concatenate([x for _ in range(n)], axis=0)

        Z = (self._rng.uniform(size=(n,)) - 0.5) * 2
        one_dim_X = self._beta @ X.T
        Z = Z * np.pi / one_dim_X

        phi = (self._rng.uniform(size=(n,))) * (2 * np.pi)
        R = 0.1 * (self._rng.uniform(size=(n,)) - 0.5) * 2
        Y1 = Z + R * np.cos(phi)
        Y2 = (-np.cos(one_dim_X * Z) + 1) / 2 + R * np.sin(phi)

        if self._nonlinear:
            Y2 += np.sin(X.mean(axis=1))

        if self._d == 2:
            Y = np.stack([Y1, Y2], axis=1)
        elif self._d == 3:
            Y3 = np.sin(Z)
            Y = np.stack([Y1, Y2, Y3], axis=1)
        elif self._d == 4:
            Y3 = np.sin(Z)
            Y4 = np.cos(np.sin(Z) + R * np.sin(phi) * np.cos(phi))
            Y = np.stack([Y1, Y2, Y3, Y4], axis=1)
        else:
            raise NotImplementedError("Cond banana is implemented only for d=2,3,4.")

        return X, Y

    def _make_beta(self) -> Array:
        beta = self._rng.uniform(low=0, high=1, size=(self.k,))
        beta /= np.linalg.norm(beta, ord=1)
        return beta


def generate_x(dataset_name, n, k):
    if dataset_name == "cond_banana" or dataset_name == "cond_quad_banana":
        X = torch.FloatTensor(n, k).uniform_(0.8, 3.2)
    else:
        assert False
    return X


def get_k_dim_banana(n, k=10, X=None, is_nonlinear=False, d=2):
    assert d in [2, 3, 4]

    beta = torch.rand(k)
    beta /= beta.norm(p=1)
    banana_beta = beta

    if X is None:
        X = generate_x("cond_banana", n, k)
    else:
        assert len(X.shape) == 2
        X = X.clone().repeat(n, 1)

    X_to_output = X

    Z = (torch.rand(n) - 0.5) * 2
    one_dim_x = banana_beta @ X.T
    Z = Z * np.pi / one_dim_x

    phi = torch.rand(n) * (2 * np.pi)
    R = 0.1 * (torch.rand(n) - 0.5) * 2
    Y1 = Z + R * torch.cos(phi)
    Y2 = (-torch.cos(one_dim_x * Z) + 1) / 2 + R * torch.sin(phi)

    if is_nonlinear:
        tmp_X = X.clone()
        if len(X.shape) == 1:
            tmp_X = tmp_X.reshape(1, len(tmp_X))
        Y2 += torch.sin(tmp_X.mean(dim=1))

    decomposed = torch.cat([R.unsqueeze(-1), phi.unsqueeze(-1), Z.unsqueeze(-1)], dim=1)
    if d == 2:
        return Y1, Y2, X_to_output, decomposed
    elif d == 3:
        Y3 = torch.sin(Z)
        return Y1, Y2, Y3, X_to_output, decomposed
    elif d == 4:
        Y4 = torch.sin(Z)
        Y3 = torch.cos(torch.sin(Z)) + R * torch.sin(phi) * torch.cos(phi)
        return Y1, Y2, Y3, Y4, X_to_output, decomposed


def get_syn_data(
    dataset_name, get_decomposed=False, k=1, is_nonlinear=False, X=None, n=None
):
    if n is None:
        if is_nonlinear:
            n = 20000
        elif k < 5:
            n = 20000
        elif k < 25:
            n = 20000
        elif k <= 60:
            n = 80000
        else:
            n = 100000

    if dataset_name == "banana":
        T = (torch.rand(n) - 0.5) * 2
        phi = torch.rand(n) * (2 * np.pi)
        Z = torch.rand(n)
        R = 0.2 * Z * (1 + (1 - T.abs()) / 2)
        Y1 = T + R * torch.cos(phi)
        Y2 = T**2 + R * torch.sin(phi)
        decomposed = torch.cat(
            [T.unsqueeze(-1), phi.unsqueeze(-1), R.unsqueeze(-1)], dim=1
        )
        Ys = [Y1, Y2]
    elif dataset_name == "cond_banana" or dataset_name == "sin_banana":
        Y1, Y2, X, decomposed = get_k_dim_banana(n, k=k, is_nonlinear=is_nonlinear, X=X)
        Ys = [Y1, Y2]
    elif dataset_name == "cond_triple_banana":
        Y1, Y2, Y3, X, decomposed = get_k_dim_banana(
            n, k=k, is_nonlinear=is_nonlinear, d=3, X=X
        )
        Ys = [Y1, Y2, Y3]
    elif dataset_name == "cond_quad_banana":
        Y1, Y2, Y3, Y4, X, decomposed = get_k_dim_banana(
            n, k=k, is_nonlinear=is_nonlinear, d=4, X=X
        )
        Ys = [Y1, Y2, Y3, Y4]
    else:
        assert False

    Y = torch.cat([y.reshape(len(y), 1) for y in Ys], dim=1)

    if get_decomposed:
        return Y, X, decomposed

    return Y, X



from typing import Tuple, Optional

import numpy as np
import pylab as pl
from numpy import ndarray
from numpy import ndarray as Array

from experiments import EXPERIMENTS_DATA_DIR
from experiments.datasets.base import DataProvider


def generate_heart() -> Tuple[ndarray, ndarray]:
    """
    Generates independent X and Y.
    X is sampled i.i.d from a uniform distribution [0, 1]
    Y is an r.v whose distribution is heart-shaped.
    :return: X, Y
    """
    # Load data from image
    image = 1 - pl.imread(EXPERIMENTS_DATA_DIR / "heart.png")[:, :, 2]
    # image = image[::2, ::2]
    image = image / np.sum(image)
    idces = image.nonzero()
    Y = np.zeros([len(idces[0]), 2])
    Y[:, 0] = idces[0] / idces[0].max()
    Y[:, 1] = idces[1] / idces[1].max()

    rng = np.random.default_rng(None)
    X = rng.uniform(size=(Y.shape[0], 1))
    return X, Y


def generate_star() -> Tuple[ndarray, ndarray]:
    """
    Generates independent X and Y.
    X is sampled i.i.d from a uniform distribution [0, 1]
    Y is an r.v whose distribution is star-shaped.
    :return: X, Y
    """
    image = 1 - pl.imread(EXPERIMENTS_DATA_DIR / "star.jpg") / 255
    # image = image[::7, ::7]
    image = image / np.sum(image)
    idces = image.nonzero()
    Y = np.zeros([len(idces[0]), 2])
    Y[:, 0] = idces[0] / idces[0].max()
    Y[:, 1] = idces[1] / idces[1].max()
    rng = np.random.default_rng(None)
    X = rng.uniform(size=(Y.shape[0], 1))
    return X, Y


def _rotation_matrix(theta: float) -> ndarray:
    return np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])


class ImageDataProvider(DataProvider):
    """
    Generates (X, Y) where Y is sampled from an images, rotated by X angles.
    """

    def __init__(
        self,
        image_data: ndarray,
        initial_rotation_deg=0,
        noise_std=0.0,
        x_max_deg=90.0,
        x_discrete=False,
        seed=42,
    ):
        Y = image_data
        # Zero mean
        Y -= np.mean(Y, axis=0)

        # Rotate
        rot_theta = _rotation_matrix(np.deg2rad(initial_rotation_deg))
        Y = Y @ rot_theta

        self._Y_gt = Y
        self._x_high = x_max_deg
        self._x_discrete = x_discrete
        self._noise_std = noise_std
        self._seed = seed
        self._rng = np.random.default_rng(seed)

    @property
    def k(self) -> int:
        return 1

    @property
    def d(self) -> int:
        return 2

    def sample_x(self, n: int) -> Array:
        if self._x_discrete:
            return self._rng.integers(0, self._x_high + 1, size=(n,)).reshape(n, -1)
        else:
            return self._rng.uniform(0, self._x_high, size=(n,)).reshape(n, -1)

    def sample(self, n: int, x: Optional[Array] = None) -> Tuple[Array, Array]:
        if x is None:
            X = self.sample_x(n)
            R = np.stack([_rotation_matrix(np.deg2rad(-x).item()) for x in X], axis=0)
        else:
            x = np.reshape(x, (1, -1))
            assert x.shape[1] == self.k
            X = np.concatenate([x for _ in range(n)], axis=0)
            R = np.stack([_rotation_matrix(np.deg2rad(-x).item())] * n, axis=0)

        N = len(self._Y_gt)
        sample_idx = self._rng.integers(0, N, size=(n,))

        Y_sampled = self._Y_gt[sample_idx]

        # Rotate samples by the corresponding x degrees
        Y_sampled = np.einsum("bji,bi -> bj", R, Y_sampled)

        # Additive noise
        eta = self._rng.normal(size=Y_sampled.shape) * self._noise_std
        Y_sampled += eta

        return X, Y_sampled


class HeartDataProvider(ImageDataProvider):
    def __init__(
        self,
        initial_rotation_deg=0,
        noise_std=0.0,
        x_max_deg=10.0,
        x_discrete=False,
        seed=42,
    ):
        Y = generate_heart()
        super().__init__(
            image_data=generate_heart()[1],
            initial_rotation_deg=initial_rotation_deg,
            noise_std=noise_std,
            x_max_deg=x_max_deg,
            x_discrete=x_discrete,
            seed=seed,
        )


class StarDataProvider(ImageDataProvider):
    def __init__(
        self,
        initial_rotation_deg=0,
        noise_std=0.0,
        x_max_deg=10.0,
        x_discrete=False,
        seed=42,
    ):
        Y = generate_heart()
        super().__init__(
            image_data=generate_star()[1],
            initial_rotation_deg=initial_rotation_deg,
            noise_std=noise_std,
            x_max_deg=x_max_deg,
            x_discrete=x_discrete,
            seed=seed,
        )
